# _*_ coding: utf-8 _*_
#
# hspylib v0.11.172
#
# Package: main.hspylib.modules.application
"""Package initialization."""

__all__ = [
  'application', 
  'argparse', 
  'exit_hooks', 
  'exit_status', 
  'parser_action', 
  'version'
]
__version__ = '0.11.172'



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HsPyLib
   @package: hspylib.modules.cli.vt100
      @file: vt_codes.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2023, HsPyLib team
"""
from hspylib.core.enums.enumeration import Enumeration
from hspylib.core.preconditions import check_not_none, check_state
from hspylib.modules.cli.vt100.vt_100 import Vt100
from typing import Callable, Optional

import os
import re


class VtCode(Enumeration):
    """VT-100 escape codes
    Ref.: http://domoticx.com/terminal-codes-ansivt100/
    """

    # fmt: off

    CSV = Vt100.save_cursor()               # ^[7 -> Save cursor position and attributes
    CRE = Vt100.restore_cursor()            # ^[8 -> Restore cursor position and attributes
    RIS = Vt100.reset()                     # ^[c -> Reset terminal to initial state

    SAW = Vt100.set_auto_wrap(True)         # ^[?7h  -> Set auto-wrap mode
    UAW = Vt100.set_auto_wrap(False)        # ^[?7l  -> Unset auto-wrap mode
    SSC = Vt100.set_show_cursor(True)       # ^[?25h -> Set show cursor
    USC = Vt100.set_show_cursor(False)      # ^[?25l -> Unset show cursor

    ED0 = Vt100.clear_screen()              # ^[[J  -> Clear screen from cursor down
    ED1 = Vt100.clear_screen(1)             # ^[[1J -> Clear screen from cursor up
    ED2 = Vt100.clear_screen(2)             # ^[[2J -> Clear entire screen

    EL0 = Vt100.clear_line()                # ^[[K  -> Clear line from cursor right
    EL1 = Vt100.clear_line(1)               # ^[[1K -> Clear line from cursor left
    EL2 = Vt100.clear_line(2)               # ^[[2K -> Clear entire line

    HOM = Vt100.set_cursor_pos()            # ^[[H  -> Move cursor to upper left corner

    SCA = Vt100.alternate_screen()          # ^[?1049h -> Switch to alternate screen buffer
    SCM = Vt100.alternate_screen(False)     # ^[?1049l -> Switch to main screen buffer

    EOL = os.linesep                        # Line separator (eol)

    # For all commands that take arguments, we need to add to this map, so we can call it.
    # The following entry values must defined as auto(), so they can be invoked as Callable

    MOD = "MOD"  # ^[[<m1;m2;m3>m  -> Set terminal modes
    CUP = "CUP"  # ^[[<v>;<h>H     -> Move cursor to screen location <v,h>
    CUU = "CUU"  # ^[[<n>A         -> Move cursor up n lines
    CUD = "CUD"  # ^[[<n>B         -> Move cursor down n lines
    CUF = "CUF"  # ^[[<n>C         -> Move cursor right n lines
    CUB = "CUB"  # ^[[<n>D         -> Move cursor left n lines

    # fmt: on

    @staticmethod
    def _vt100_fnc(command: str) -> Optional[Callable]:
        """Return the corresponding VT100 function specified by command."""
        fnc = None
        match command:
            case "MOD":
                fnc = Vt100.mode
            case "CUP":
                fnc = Vt100.set_cursor_pos
            case "CUU":
                fnc = Vt100.cursor_move_up
            case "CUD":
                fnc = Vt100.cursor_move_down
            case "CUF":
                fnc = Vt100.cursor_move_forward
            case "CUB":
                fnc = Vt100.cursor_move_backward
        return fnc

    @classmethod
    def decode(cls, input_string: str) -> str:
        """Decode the string into a VT_CODE enum."""
        check_not_none(input_string)
        commands = re.findall(r"%([a-zA-Z0-9]+)(\([0-9]+(;[0-9]+)*\))?%", input_string)
        for cmd in commands:
            if (mnemonic := cmd[0]) in VtCode.names():
                args = cmd[1][1:-1] if cmd[1] else None
                if args:  # If has args we need to invoke the vt100 function
                    if fnc := VtCode.value_of(mnemonic).__call__:
                        input_string = input_string.replace(f"%{mnemonic + cmd[1]}%", fnc(args) if fnc else "")
                else:
                    input_string = input_string.replace(f"%{mnemonic}%", str(VtCode.value_of(mnemonic).value))

        return input_string

    def __call__(self, *args, **kwargs) -> str:
        fn = VtCode._vt100_fnc(self.name)
        check_state(fn and callable(fn), f"Could not find a proper function for {self.name}")
        return fn(*args)

    def __str__(self) -> str:
        return str(self.value)

    @property
    def code(self) -> str:
        """Decode the string into a VT_CODE enum"""
        return str(self)

    @property
    def placeholder(self) -> str:
        """Decode the string into a VT_CODE enum"""
        return f"%{self.name}%"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HsPyLib
   @package: hspylib.modules.application
      @file: parser_action.py
   @created: hu, 14 Feb 2022
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2023, HsPyLib team
"""
from hspylib.core.enums.enumeration import Enumeration


class ParserAction(Enumeration):
    """TODO"""

    # fmt: off

    # This stores a list, and appends each argument value to the list.
    APPEND = 'append'

    # This stores a list, and appends the value specified by the const keyword argument to the list.
    APPEND_CONST = 'append_const'

    # This counts the number of times a keyword argument occurs.
    COUNT = 'count'

    # This just stores the argument’s value.
    STORE = 'store'

    # This stores the value specified by the const keyword argument.
    STORE_CONST = 'store_const'

    # These is a special cases of 'store_const' used for storing True.
    STORE_TRUE = 'store_true'

    # These is a special cases of 'store_const' used for storing False.
    STORE_FALSE = 'store_false'

    # This expects a version= keyword argument, prints version information and exits when invoked.
    VERSION = 'version'

    # fmt: on

    def __str__(self):
        return f"{self.value}"

    def __repr__(self):
        return str(self)



# _*_ coding: utf-8 _*_
#
# hspylib v0.11.172
#
# Package: main.hspylib.core
"""Package initialization."""

__all__ = [
  'collection_filter', 
  'config', 
  'constants', 
  'decorator', 
  'enums', 
  'exception', 
  'metaclass', 
  'namespace', 
  'preconditions', 
  'tools', 
  'zoned_datetime'
]
__version__ = '0.11.172'



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HsPyLib
   @package: hspylib.modules.fetch
      @file: http_response.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2023, HsPyLib team
"""
from hspylib.core.enums.charset import Charset
from hspylib.core.enums.content_type import ContentType
from hspylib.core.enums.http_code import HttpCode
from hspylib.core.enums.http_method import HttpMethod
from requests.models import CaseInsensitiveDict, Response
from typing import Optional


class HttpResponse:
    """TODO"""

    @staticmethod
    def of(response: Response) -> "HttpResponse":
        """TODO"""
        return HttpResponse(
            HttpMethod.value_of(response.request.method),
            response.url,
            HttpCode.of(response.status_code),
            response.headers,
            response.text,
            Charset.of_value(response.encoding.lower()) if response.encoding else Charset.UTF_8,
        )

    def __init__(
        self,
        method: HttpMethod,
        url: str,
        status_code: HttpCode,
        headers: CaseInsensitiveDict = None,
        body: Optional[str] = None,
        encoding: Charset = Charset.UTF_8,
        content_type=ContentType.APPLICATION_JSON,
    ):
        self.method = method
        self.url = url
        self.status_code = status_code
        self.body = body
        self.headers = headers
        self.encoding = encoding
        self.content_type = content_type
        if self.content_type:
            self.content_type.charset = self.encoding

    def __repr__(self):
        return f"{repr(self.status_code)} {self.url}"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HsPyLib
   @package: hspylib.core.tools
      @file: commons.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2023, HsPyLib team
"""
from datetime import timedelta
from hspylib.core.constants import TRUE_VALUES
from hspylib.core.enums.charset import Charset
from hspylib.core.preconditions import check_argument, check_not_none
from hspylib.modules.cli.vt100.vt_code import VtCode
from hspylib.modules.cli.vt100.vt_color import VtColor
from typing import Any, Callable, Iterable, Optional, Set, Tuple, Type

import inspect
import logging as log
import os
import pathlib
import signal
import sys

# pylint: disable=consider-using-f-string
FILE_LOG_FMT = "{}\t{} {} {} {} {} ".format(
    "%(asctime)s",
    "%(levelname)-5.5s",
    "%(filename)s::",
    "%(message)s",
    "%(funcName)s(@Line:%(lineno)d) -",
    "%(threadName)-12.12s",
)

CONSOLE_LOG_FMT = "{}\t{} {} {} {} ".format(
    "%(levelname)-5.5s", "%(message)s", "%(funcName)s(@Line:%(lineno)d)", "%(asctime)s -", "%(threadName)-12.12s"
)


def log_init(
    filename: str = "",
    filemode: str = "a",
    level: int = log.DEBUG,
    log_format: str = FILE_LOG_FMT,
    clear_handlers: bool = True,
    console_enable: bool = False,
    file_enable: bool = True,
) -> bool:
    """Initialize the system logger"""

    # if someone tried to log something before log_init is called, Python creates a default handler that is going to
    # mess our logs. Remove handlers if there is any.
    root, handlers = log.getLogger(), set()

    if clear_handlers:
        if root.handlers:
            for handler in root.handlers:
                handler.close()
                root.removeHandler(handler)

    if file_enable:
        if not os.path.exists(filename):
            touch_file(filename)
        touch_file(filename)
        file_formatter = log.Formatter(log_format)
        file_handler = log.FileHandler(filename=filename, mode=filemode)
        file_handler.setFormatter(file_formatter)
        handlers.add(file_handler)

    if console_enable or (file_enable and not os.path.exists(filename)):
        console_formatter = log.Formatter(CONSOLE_LOG_FMT, "%Y-%m-%d %H:%M:%S")
        console_handler = log.StreamHandler(sys.stdout)
        console_handler.setFormatter(console_formatter)
        handlers.add(console_handler)

    log.basicConfig(level=level, handlers=handlers)

    return file_enable if os.path.exists(filename or "") else console_enable


def is_debugging() -> bool:
    """Whether the program is running under debug mode."""
    for frame in inspect.stack():
        if frame[1].endswith("pydevd.py"):
            return True
    return False


def dirname(filepath: str) -> str:
    """Retrieve the directory of the specified filepath"""
    return os.path.dirname(os.path.realpath(filepath))


def run_dir() -> str:
    """Retrieve the application's root directory"""
    return sys.path[0]


def get_path(filepath: str) -> pathlib.Path:
    """Retrieve the Path of the file"""
    return pathlib.Path(filepath).parent


def sysout(*strings: str, end: str = os.linesep) -> None:
    """Print the unicode input_string decoding vt100 placeholders
    :param strings: values to be printed to sys.stdout
    :param end: string appended after the last value, default a newline
    """
    if strings:

        def sysout_format(text: str) -> str:
            return VtColor.colorize(VtCode.decode(text or ""))

        list(map_many(strings, sysout_format, lambda s: print(s, file=sys.stdout, flush=True, end="")))
        print("", file=sys.stdout, flush=True, end=end)


def syserr(*strings: Any, end: str = os.linesep) -> None:
    """Print the unicode input_string decoding vt100 placeholders
    :param strings: values to be printed to sys.stderr
    :param end: string appended after the last value, default a newline
    """
    if strings:

        def syserr_format(text: Any) -> str:
            plain_text = VtColor.strip_colors(str(text or ""))
            return VtColor.colorize(VtCode.decode(f"%RED%{plain_text}%NC%"))

        list(map_many(strings, syserr_format, lambda s: print(s, file=sys.stderr, flush=True, end="")))
        print("", file=sys.stdout, flush=True, end=end)


def hook_exit_signals(handler: Callable) -> None:
    """Hook common exit signals and set proper handlers for them
    :param handler a Callable to handle the exit signals
    """
    check_argument(handler is not None and isinstance(handler, Callable))
    signal.signal(signal.SIGINT, handler)
    signal.signal(signal.SIGTERM, handler)
    signal.signal(signal.SIGHUP, handler)
    signal.signal(signal.SIGABRT, handler)


def class_attribute_names(clazz: Type) -> Optional[Tuple]:
    """Retrieve all attribute names of the class
    :param clazz: The class to retrieve the attribute names
    """
    return tuple(vars(clazz()).keys()) if clazz else None


def class_attribute_values(instance: dict) -> Optional[Tuple]:
    """Retrieve all attribute values of the class
    :param instance: The class to retrieve the attribute values
    """
    return tuple(instance.values()) if instance else None


def str_to_bool(string: str, true_values: Set[str] = None) -> bool:
    """Convert a string to boolean
    :param string: The string to be converted
    :param true_values: The list of strings that will become True value
    """
    return string is not None and string.lower() in (true_values or TRUE_VALUES)


def map_many(iterable: Iterable, function: Callable, *functions) -> Optional[map]:
    """Maps multiple functions to the same iterable
    :param iterable The iterable to map
    :param function The first function to be mapped
    :param functions the other functions to be mapped
    """
    check_not_none((iterable, function))
    if functions:
        # Recursively map all functions
        return map_many(map(function, iterable), *functions)

    # Map the single function
    return map(function, iterable)


def safe_delete_file(filename: str, on_not_found_except: bool = False) -> bool:
    """Delete the file specified by filename. If the file is not found, raises an exception if on_not_found_except is
    True; otherwise return False.
    :param filename: the name of the file to be checked
    :param on_not_found_except: boolean parameter to raise an exception if the file is not found.
    """
    if os.path.exists(filename):
        os.remove(filename)
    else:
        if on_not_found_except:
            raise FileNotFoundError(f"File was not found on the system: {filename}")
        return False

    return True


def file_is_not_empty(filename: str) -> bool:
    """Check whether the file is empty or not.
    :param filename: the name of the file to be modified. If the file does not exist; return False
    """
    return os.path.exists(filename) and os.stat(filename).st_size > 0


def touch_file(filename: str, encoding: str = Charset.UTF_8.val) -> None:
    """Change file modification time
    :param filename: the name of the file to be modified
    :param encoding: the file encoding
    """
    with open(filename, "a", encoding=encoding):
        os.utime(filename, None)


def human_readable_bytes(size_in_bytes: int) -> Tuple[str, str]:
    """Return a Human readable bytes and unit
    :param size_in_bytes: the size to be formatted
    """

    byte_size = float(size_in_bytes)
    kb, mb, gb, tb = 2**10, 2**20, 2**30, 2**40

    if 0 <= byte_size <= kb:
        ret_val = f"{byte_size:3.2f}"
        ret_unit = "[B]"
    elif kb < byte_size <= mb:
        ret_val = f"{byte_size / kb:3.2f}"
        ret_unit = "[Kb]"
    elif mb < byte_size <= gb:
        ret_val = f"{byte_size / mb:3.2f}"
        ret_unit = "[Mb]"
    elif gb < byte_size <= tb:
        ret_val = f"{byte_size / gb:3.2f}"
        ret_unit = "[Gb]"
    else:
        ret_val = f"{byte_size / tb:3.2f}"
        ret_unit = "[Tb]"

    return ret_val, ret_unit


def human_readable_time(time_microseconds: int) -> str:
    """Return a Human readable formatted time using format: HH:MM:SS.uuuuuu
    :param time_microseconds: the time to be formatted
    """
    delta = timedelta(microseconds=time_microseconds)
    total_seconds = delta.seconds
    seconds = total_seconds % 60
    minutes = total_seconds / 60 % 60
    hours = total_seconds / 3600
    microseconds = delta.microseconds
    return f"{hours:02d}:{minutes:02d}:{seconds:02d}.{microseconds:06d}"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HsPyLib
   @package: hspylib.modules.cli.vt100
      @file: vt_color.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2023, HsPyLib team
"""
from enum import auto
from hspylib.core.enums.enumeration import Enumeration
from hspylib.modules.cli.vt100.vt_100 import Vt100


# pylint: disable=multiple-statements
# @composable
class VtColor(Enumeration):
    """
    Ref.: https://en.wikipedia.org/wiki/ANSI_escape_code#SGR
    """

    # fmt: off
    _CUSTOM     = auto()
    NC          = Vt100.mode('0;0;0')       ; BLACK      = Vt100.mode('30')
    RED         = Vt100.mode('31')          ; BG_RED     = Vt100.mode('41')
    GREEN       = Vt100.mode('32')          ; BG_GREEN   = Vt100.mode('42')
    YELLOW      = Vt100.mode('93')          ; BG_YELLOW  = Vt100.mode('43')
    BLUE        = Vt100.mode('34')          ; BG_BLUE    = Vt100.mode('44')
    PURPLE      = Vt100.mode('35')          ; BG_PURPLE  = Vt100.mode('45')
    CYAN        = Vt100.mode('36')          ; BG_CYAN    = Vt100.mode('46')
    GRAY        = Vt100.mode('38;5;8')      ; BG_GRAY    = Vt100.mode('47')
    ORANGE      = Vt100.mode('38;5;202')    ; BG_ORANGE  = Vt100.mode('48;5;202')
    VIOLET      = Vt100.mode('95')          ; BG_VIOLET  = Vt100.mode('48;5;90')
    WHITE       = Vt100.mode('97')          ; BG_WHITE   = Vt100.mode('48;5;255')
    # fmt: on

    @classmethod
    def colorize(cls, input_string: str) -> str:
        colorized = input_string
        for color in cls.names():
            colorized = cls._replace_all(colorized, color)
        return colorized

    @classmethod
    def strip_colors(cls, input_string: str) -> str:
        plain = input_string
        for color in cls.names():
            plain = plain.replace(f"%{color}%", "")
        return plain

    @classmethod
    def _replace_all(cls, input_string: str, color_name: str) -> str:
        color = VtColor.value_of(color_name.upper())
        return input_string.replace(color.placeholder, color.code)

    @property
    def code(self) -> str:
        return str(self.value)

    @property
    def placeholder(self) -> str:
        return f"%{self.name}%"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HsPyLib
   @package: hspylib.core.config
      @file: parser_factory.py
   @created: Wed, 19 May 2023
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2023, HsPyLib team
"""
from abc import ABC
from configparser import ConfigParser
from functools import partial
from hspylib.core.tools.dict_tools import flatten_dict
from typing import Any, Callable, Dict, TextIO, TypeVar

import os
import re
import toml
import yaml

PROPERTIES = TypeVar("PROPERTIES", bound=[Dict[str, Any]])


class ParserFactory(ABC):
    class PropertyParser:
        def __init__(self, parser: Callable[[TextIO], PROPERTIES]):
            self._parser = parser

        def parse(self, file_handler: TextIO) -> PROPERTIES:
            return self._parser(file_handler)

    @staticmethod
    def _read_properties(file_handler: TextIO) -> PROPERTIES:
        """Reads properties from properties file (key and element pairs) from the input list."""
        all_lines = list(map(str.strip, filter(None, file_handler.readlines())))
        # fmt: off
        return {
            p[0].strip(): p[1].strip()
            for p in [
                p.split("=", 1) for p in list(
                    filter(lambda l: re.match(r"[a-zA-Z]([.\\-]|\w)* *= *.+", l), all_lines)
                )
            ]
        }
        # fmt: off

    @staticmethod
    def _read_cfg_or_ini(file_handler: TextIO) -> PROPERTIES:
        """Reads properties from a cfg or ini file (key and element pairs) from the input list."""
        all_lines = list(map(str.strip, filter(None, file_handler.readlines())))
        string = os.linesep.join(all_lines)
        all_cfgs = {}
        cfg = ConfigParser()
        cfg.read_string(string)
        for section in cfg.sections():
            all_cfgs.update(dict(cfg.items(section)))
        return all_cfgs

    @staticmethod
    def _read_yaml(file_handler: TextIO) -> PROPERTIES:
        """Reads properties from a yaml file (key and element pairs) from the input list."""
        return flatten_dict(yaml.safe_load(file_handler))

    @staticmethod
    def _read_toml(file_handler: TextIO) -> PROPERTIES:
        """Reads properties from a toml file (key and element pairs) from the input list."""
        return flatten_dict(toml.load(file_handler))

    @classmethod
    def create(cls, file_extension: str) -> PropertyParser:
        if file_extension in [".ini", ".cfg"]:
            parser = partial(cls._read_cfg_or_ini)
        elif file_extension == ".properties":
            parser = partial(cls._read_properties)
        elif file_extension in [".yml", ".yaml"]:
            parser = partial(cls._read_yaml)
        elif file_extension == ".toml":
            parser = partial(cls._read_toml)
        else:
            raise NotImplementedError(f"Extension {file_extension} is not supported")

        return cls.PropertyParser(parser)



# _*_ coding: utf-8 _*_
#
# hspylib v0.11.172
#
# Package: main.hspylib.core.tools
"""Package initialization."""

__all__ = [
  'commons', 
  'dict_tools', 
  'git_tools', 
  'json_path', 
  'text_tools', 
  'validator'
]
__version__ = '0.11.172'



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HsPyLib
   @package: hspylib.modules.application
      @file: version.py
   @created: Thu, 14 Feb 2022
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2023, HsPyLib team
"""
from hspylib.core.enums.charset import Charset
from hspylib.core.preconditions import check_argument
from pathlib import Path
from typing import Union

import logging as log
import os
import re


class Version:
    """
    Ref.: https://semver.org/
    """

    @staticmethod
    def initial() -> "Version":
        return Version(0, 1, 0)

    @staticmethod
    def unversioned() -> "Version":
        return Version(-1, -1, -1)

    @staticmethod
    def load(filename: str = ".version", load_dir: Union[str, Path] = os.getcwd()) -> "Version":
        """Load a version from file."""
        filepath = f"{str(load_dir)}/{filename}"
        if not os.path.exists(filepath):
            log.warning('File "%s" does not exist. Could not fetch application version from %s', filepath, load_dir)
            return Version.unversioned()
        return Version._read(filepath)

    @staticmethod
    def _read(filepath: str = ".version") -> "Version":
        """Retrieve the version from the version file."""
        log.debug("Reading version from %s", filepath)
        with open(filepath, encoding=Charset.UTF_8.val) as fh_version:
            ver_str = fh_version.read().strip()
            if mat := re.search(r"v?((\d+)\.(\d+)\.(\d+)(.*))", ver_str):
                delimiters = re.findall(r"[-_ ]", ver_str)
                extras = re.split(r"[-_ ]", ver_str)[1:]
                ext = list(map("".join, zip(delimiters, extras)))
                return Version(mat.group(2), mat.group(3), mat.group(4), *ext)
            return Version.unversioned()

    def __init__(self, *args) -> None:
        """Application version. Additional labels for pre-release and build metadata are available as extensions
        to the MAJOR.MINOR.PATCH format:
            - MAJOR version when you make incompatible API changes
            - MINOR version when you add functionality in a backwards compatible manner
            - PATCH version when you make backwards compatible bug fixes
        :param version: The version number according to the following: MAJOR.MINOR.PATCH
        """
        check_argument(args and len(args) >= 3, "Version must have at least MAJOR.MINOR.PATCH numbers")
        self._major = args[0]
        self._minor = args[1]
        self._patch = args[2]
        self._ext = "".join(args[3:])
        self._version_string = f"{'.'.join(map(str, args[:3]))}{self._ext}"

    def __str__(self) -> str:
        return self._version_string

    def __repr__(self) -> str:
        return str(self)

    @property
    def version(self) -> str:
        """TODO"""
        return f"v{str(self)}"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HsPyLib
   @package: hspylib.modules.cache
      @file: ttl_keyring_be.py
   @created: Tue, 4 Oct 2022
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2023, HsPyLib team
"""
from datetime import datetime
from hspylib.core.preconditions import check_not_none
from hspylib.core.zoned_datetime import now_ms
from hspylib.modules.security.security import b64_decode, b64_encode
from keyring.backends.chainer import ChainerBackend
from keyring.errors import PasswordDeleteError, PasswordSetError
from typing import Any, Callable, Optional

import json
import keyring

PWD_EXPIRED_CB = Callable[[str], Any]


class TTLKeyringBE(ChainerBackend):
    priority = 1

    def __init__(self, ttl_minutes: int = 15, ttl_seconds: int = 0, cb_expired: PWD_EXPIRED_CB = None) -> None:
        super().__init__()
        self._ttl = ttl_minutes, ttl_seconds
        self._cb_expired = cb_expired

    def set_password(self, service: str, username: str, password: str) -> None:
        """Set password for the username of the service."""
        try:
            check_not_none(service, username, password)
            expires_sec = now_ms() + (self._ttl[0] * 60 + self._ttl[1])
            passwd_obj = {"sn": service, "un": username, "pw": password, "ttl": expires_sec}
            b64_pwd = b64_encode(json.dumps(passwd_obj))
            super().set_password(service, username, b64_pwd)
        except PasswordSetError:
            pass  # it does not matter if the password set failed.

    def get_password(self, service: str, username: str) -> Optional[str]:
        """Get password of the username for the service."""
        check_not_none(service, username)
        if b64_pwd := super().get_password(service, username):
            passwd_str = b64_decode(b64_pwd)
            passwd_obj = json.loads(passwd_str)
            dt_object = datetime.fromtimestamp(passwd_obj["ttl"])
            if now_ms() - dt_object.timestamp() > 0:
                keyring.delete_password(service, username)
                if self._cb_expired:
                    self._cb_expired(passwd_obj["pw"])
            else:
                return passwd_obj["pw"]

        return None

    def delete_password(self, service: str, username: str) -> None:
        """Delete the password for the username of the service."""
        check_not_none(service, username)
        try:
            super().delete_password(service, username)
        except PasswordDeleteError:
            pass  # it does not matter if the password does not exist.
